(* ::Package:: *)

(*
 *******************************************************
	Package: IDERKSolve
	Author:  Janos Karsai
	Address: Dept. Medical Informatics,
		 University of Szeged
	         H-6720, Szeged, Hungary
	E-mail:  karsai@dmi.u-szeged.hu
 *******************************************************
*)
(*********************************************************************
       Based on the ODE-solver package from
        Roman E. Maeder: Programming in Mathematica,
        Second Edition, Addison-Wesley, 1991.
 *********************************************************************)
(* Impulsive system solver using Runge-Kutta method. 2001. Janos Karsai  *)
(* Upgraded to Version 6: 2008. Janos Karsai  *)


(*
*******************************************************
* Development is supported by 
* Foundation for the Hungarian Higher Education Research, Grant No. 776/94
* Foundation for the Development of Higher Education Grant No. FEFA-IV 1608/4.
* Hungarian Science Research Foundation OTKA T-034275 
* Hungarian Ministry of Education: Project o. 179/1996
* Pro Renovanda Cultura Hungarie: "Science in Education"
* Hungarian Ministry of Education: Higher Education Textbook Grant, 2001
* Hungarian National Foundation for Scientific Research Grant No. T 049516
* INTERREG IIIA Community Initiative Programme Grant No. HUSER0602/066S.
*******************************************************
*)



BeginPackage["IDERKSolve`"]

IDERKSolve::usage =
	"IDERKSolve`[
		{rhs1,rhs2,...,rhsK}, 
		{{surface1,Impulse1,stepoption1},...,
		 {surfacek,Impulsek,stepoptionk}},
		{y1,y2,...,yK}, {{y01,y02,...y0K},..,}, {t,t0,t1,dt}]
integrates the ODE with right-hand-side {rhs1,rhs2,...,rhsK} containing the variables 
t,y1,y2,...,yK with several inital values {y01,y02,..}... using a RungeKutta 
method. Functions surfacei are scalar valued expressions of the variables t,y1,y2,...,yK.
Functions Impulsei are K dimensional vectors of expresssions of t,y1,y2,...,yK.
The solutions have impulses Impulsei at the points satisfying surfacei==0.
Integration proceeds stepwise, stepsize is dt.
The ODE part is solved using a Runge-Kutta method implemented by  Roman E. Maeder. 
If the i'th function surfacei changes the sign moving from the time t_n to t_{n+1},
then the i-th impulse function impulsei is applied at t_n (with setting stepoption=0) 
or at t_{n+1} (with setting stepoption=1).  

See some examples in the source code.
The functions Impulse-i MUST be given in lists!";

Begin["`Private`"]

Coordinate[data_,n_]:= Map[{#[[1]],#[[n]]}&,data];

RKImpStep[f_, Impulse_, y_, y0_, dt_] :=
	Module[{ k1, k2, k3, k4 },
		k1 = dt N[ f /. Thread[y -> y0] ];
		k2 = dt N[ f /. Thread[y -> y0 + k1/2] ];
		k3 = dt N[ f /. Thread[y -> y0 + k2/2] ];
		k4 = dt N[ f /. Thread[y -> y0 + k3] ];
	Apply[
	Which[##1, True, y0  + (k1 + 2 k2 + 2 k3 + k4)/6]&,
		Flatten[
		Map[
		{Apply[Times, 
		(# [[1]]/. Thread[y -> y0])*
        (#[[1]] /. Thread[y ->(y0+(k1 + 2 k2 + 2 k3 + k4)/6) ])] <= 0., 
        (#[[2]] /. Thread[y -> (y0  + #[[3]]*(k1 + 2 k2 + 2 k3 + k4)/6)])}&,
      Impulse]
		,1]
     ]
   ]


IDERKSolve[f_List, Impulse_, y_List, y0_List, {t1_, dt_}] :=Module[{},
 	NestList[RKImpStep[f, Impulse, y, #1, N[dt]]& , N[y0], 
    Round[N[t1/dt]]] /; Length[f] === Length[y] === Length[y0]
    ]

IDERKSolve[f_List,Impulse_, y_List, y0_List, {t_, t0_, t1_, dt_}] :=
		Module[{res,init,ii},
		If[Length[f]!= Length[y],Return[]];
		init=Which[Depth[y] === Depth[y0],{Prepend[y0, t0]},
		Depth[y]+1 === Depth[y0],Map[Prepend[#,t0]&,y0],
		True, Return[]];
		Return[
			res = Map[
				IDERKSolve[ Prepend[f, 1],
					Map[{#[[1]],Prepend[#[[2]],t],#[[3]]}&,Impulse],  
					Prepend[y, t], #, {t1 - t0, dt}]&,init]
					]
	(* Return[
		Map[Table[Interpolation[Coordinate[#,ii+1],
		InterpolationOrder->1],{ii,1,Length[y]}]&,res]]
	*)	
	];
End[]

(* Protect[IDERKSolve] *)

EndPackage[]

(*
  *******************************************************
	EXAMPLES
  *******************************************************

Coordinate[data_,n_]:= Map[{#[[1]],#[[n]]}&,data]

************
* Example 1  *
************
xdot={y,-4};
xvar={x,y};
init={{0.9,2.2},{0.9,2.1}};
T=9;
dt=0.01;

Impulse={{x-1,{x,-1.3y},0.},{x,{x,-0.9y},0.}};

sol1=IDERKSolve[xdot,Impulse,xvar,init,{t,0,T,dt}];

ListPlot[Coordinate[sol1[[1]],2],GridLines->Automatic,PlotRange->All,
    PlotJoined->True];

ListPlot[Coordinate[sol1[[2]],2],GridLines->Automatic,PlotRange->All,
    PlotJoined->True];


************
* Example 1  *
************

xdot={0};
xvar={x};
init={{-1},{-1.5},{-0.1},{1},{1.5},{0.1}};
T=4;
dt=0.01;
Impulse={{Which[Abs[x]<1,Sin[Pi(t-x)],True,1],{Sqrt[Abs[x]]Sign[x]},1}};

sol1=IDERKSolve[xdot,Impulse,xvar,init,{t,0,T,dt}];

ListPlot[sol1[[6]],PlotJoined->True,GridLines->Automatic,PlotRange->All];

*************************************************************************
*)



