(* ::Package:: *)

(*
 *******************************************************
	Package: IDESolve
	Author:  Janos Karsai
	Address: Dept. Medical Informatics,
		 University of Szeged
	         H-6720, Szeged, Hungary
	E-mail:  karsai@dmi.u-szeged.hu
 *******************************************************
*)
(*
*******************************************************
Package for solution and visualization of impulsive 
differential equations with fixed instants of impulses

Version: Mathematica 6.0

2008.

* Development is supported by 
* Foundation for the Hungarian Higher Education Research, Grant No. 776/94
* Foundation for the Development of Higher Education Grant No. FEFA-IV 1608/4.
* Hungarian Science Research Foundation OTKA T-034275 
* Hungarian Ministry of Education: Project o. 179/1996
* Pro Renovanda Cultura Hungarie: "Science in Education"
* Hungarian Ministry of Education: Higher Education Textbook Grant, 2001
* Hungarian National Foundation for Scientific Research Grant No. T 049516
* INTERREG IIIA Community Initiative Programme Grant No. HUSER0602/066S.
*
*******************************************************
*)


BeginPackage["`IDESolve`"]; 

Search::usage = "Search[{l1,l2, ...},k,low,high] gives the 
index ind between low and high for which l[[ind]]<=k<=
l[[ind+1]]. If l[[1]]>k then 0 is returned.";

JoinInterpol::usage = "JoinInterpol[
Interpolationfunction[...],Interpolationfunction[...]]
joins the given InterpolationFunctions objects and returns
an InterpolationFunction object with argument.";
 
IcondEqn::usage = "IcondEqn[
[xvar_List,x_List,Imp_,tn_List,nn_]
gives the equations for initial values.";

IDESolve::usage = 
"IDESolve[{rhs1,rhs2,...,rhsK},{x1,x2,...,rhsK},{t1,t2,...},
Imp,{{x01,x02,...,x0K},{x11,x12,...,x1K}...},{t,t0,T},opt] solves
the ordinary impulsive system with fixed impulse times and 
with several initial conditions on the interval [t0,T]. 
NDSolve with opt options is used to solve the ordinary differential
system. 
The right hand side of the ODE part is {rsh1,rsh2,...}.
The dependent variables are {x1,x2,...}, 
Imp[n,{t1,t2,...},{x1,x2,...,xK}] is the impulse function. 
Its parameters are
	n - the index of the impulse, 
	{t1,t2,...} - the list of the impulse moments,
	{x1,x2,...,xK} - the value of the solution at the given moment.

The function gives the solution in the list of InterplatingFunction objects. 
with argument. Example is given in the source file.";


Begin["`Private`"];
  
SetAttributes[Search,HoldFirst];
Search[l_,k_,low_Integer,high_Integer]:=
	Module[{lo,mid,hi,el,ls},		
        ls = l;
		lo = Max[1,low];
        hi = Min[high,Length[ls]];
	If[Length[ls]>0,		
		If[ls[[lo]]>k,Return[0]];
		If[ls[[hi]]<k,Return[hi]];        	
        While[lo <= hi,
                If[((el=ls[[mid = Floor[ (lo + hi)/2 ]]])<=k)
 					&& (ls[[mid+1]]>=k),
                    Return[mid]];
                If[el > k, hi = mid-1];
				If[el < k, lo =mid+1];
		     ],Return[0]
        ];
		Return[lo-1/2]
	];

(*
IcondEqn
*)

IcondEqn[xvar_List,xtn_List,Imp_,tn_List, nn_]:=
Thread[Map[#[tn[[nn]]]&,xvar]==Imp[nn,tn,xtn]];

(*
JoinInterpol
*)

SetAttributes[JoinInterpol, HoldAll]; 
JoinInterpol[IP1_, IP2_] := Module[{tmp, tmp1, tmp2, it1, it2}, If[Length[IP1] == 0, Return[IP2]]; 
     If[Length[IP2] == 0, Return[IP1]]; {IP1[[0]], IP2[[0]]} /. 
      {InterpolatingFunction[{{t01_, t11_}}, {a11___, {l1_}, a21___}, {x1_}, {packedarray1_, num1_, y1_}, 
         aut1_], InterpolatingFunction[{{t02_, t12_}}, {a12___, {l2_}, a22___}, {x2_}, 
         {packedarray2_, num2_, y2_}, aut_]} :> InterpolatingFunction[{{t01, t12}}, {a11, {l1 + l2}, a21}, 
         {Join[Drop[x1, -1], {Last[x1] - Take[x1, -2] . {-1, 1}/10^9}, x2]}, 
         {packedarray1, Join[num1, Drop[num2 + Last[num1] + 2, -1]], Join[y1, y2]}, aut][IP1[[1]]]]; 


(*
 Impulsive eqn. solution (InterpolationFunction object)
*)

IDESolve0[xdot_List,xvar_List,tn_List,Imp_,icond_List,
		{t_,t0_,T_},opt___]:=
Module[{eqn,i,j,init,nn,n0,len,solv,var,tt0,tt1},
len=Length[xvar];

If[len!=Length[xdot] || len!=Length[icond],
         Print["IDESolve:Error in dimensions"];Return[]];

If[t0>T, Print["t0<T is not satisfied"]; Return[]];
var= xvar /. Map[#->#[t]&,xvar] ;
eqn=Table[xvar[[i]]'[t]==
         (xdot[[i]]/. Map[#->#[t]&, xvar]),{i,1,len}];
init=Table[xvar[[i]][t0]==icond[[i]],{i,1,len}];
nn=Floor[Search[tn,T,1,Length[tn]]];
n0=Floor[Search[tn,t0,1,Length[tn]]];


If[nn==n0,
	Return[ var /.First[
		NDSolve[Join[eqn,init],var,
		{t,t0,T},opt]]
		]];

solv= var /.First[
		NDSolve[Join[eqn,init],var,
		{t,t0,Min[tn[[n0+1]],T]},opt]];

If[T<=tn[[n0+1]],Return[solv]];
For[i=n0+1,(i<nn),i++,
	init=IcondEqn[xvar,
			solv /.t->tn[[i]],Imp,tn,i];	

solv=MapThread[JoinInterpol,
		{solv,(var /. 
		First[NDSolve[Join[eqn,init],var,
			         {t,tn[[i]],tn[[i+1]]},opt]])
		}];	
];
If[tn[[nn]]==T,Return[solv]];

init=IcondEqn[xvar,Table[solv[[j]]/.t->tn[[nn]],{j,1,len}],
			 Imp,tn,nn];	
solv=MapThread[JoinInterpol,{solv,
         (var /. First[NDSolve[Join[eqn,init],var,
		{t,tn[[nn]],T},opt]])}
 ];

Return[solv];
];

(*
 Table of Impuls. eqn. solution (table of Interpol.Funct. object)
*)

IDESolve[xdot_List,xvar_List,tn_List,Imp_,ictab_List,
			{t_,t0_,T_},opt___]:=
	Map[(IDESolve0[xdot,xvar,tn,Imp,#,{t,t0,T},opt])&,ictab];

End[];
(*
Protect[IDESolve];
*)
EndPackage[]; 

(* An Example *)
(*
p=N[2];
Imax=50;
tn=Table[n p,{n,0,Imax}];
bn=Table[{1,1},{i,0,Imax}];
Clear[ImpMult];
ImpMult[n_,tn_List,xn_List]:= MapThread[Times,{bn[[n]],xn}];
*)
(*
Additive effect: yi[tn[[k]]+0]=yi[tn[[k]]-0] + bn[[k]]*yi[tn[[k]]-0]
*)
(*
Clear[ImpAdd];
p=N[2];
Imax=50;
tn=Table[n p,{n,0,Imax}];
bn=Table[{1,1},{i,0,Imax}];
ImpAdd[n_,tn_List,xn_List]:=MapThread[Times,{(bn[[n]]+1),xn}];

*)
  
(*
Plot of impulse effect
*)

(*
PlotImp1[tn_,bn_,Icoord_,I_,I0_,IN_]:=
ParametricPlot[{tn[[I]],y},{y,0,bn[[I,Icoord]]}]

??????????

PlotImp[tn_List,Imp_,Icoord_,t0_,T_]:=
Module{n0,N}
len=Length[tn];
N=Floor[Search[tn,T,1,len]];
n0=Floor[Search[tn,t0,1,len]];
If[n0=-1,n0=1];
Plot3D[Imp[i,tn[[i]],?????????
 (Show[Graphics[#],PlotRange->All,Axes->True])& [Table[
 Line[{{tn[[i]],0},{tn[[i]],bn[[i,Icoord]]}}],{i,I0,IN,1}]]
*)

(*
 Example: linear system with impulses 
*)

(*
The right hand side and the variables
*)
(*
Clear[xdot,xvar,x,y];
xdot={y,-x+y/t};
xvar={x,y};
*)


(*
Impulse function with multiplication
*)


(*
p=N[2];
Imax=50;
tn=Table[n p,{n,0,Imax}]; (* Instants of impulses *);
bn=Table[{1,1},{i,0,Imax}];
Clear[Imp];
Imp[n_,tn_List,xn_List]:=
MapThread[Times,{bn[[n]],xn}];
*)

(*
Call IDESolve
*)

(*
icond={1,0};
t0=1.;
T=20;
sol=IDESolve[xdot,xvar,tn,Imp,icond,{t,t0,T}]
*)
(*
The output:
{InterpolatingFunction["<<>>"], 
  InterpolatingFunction["<<>>"]}
*)

(*
Call IDESolveTab 
*)

(*
ictab={{1,1},{0,1}};
soltab=IDESolveTab[xdot,xvar,tn,Imp,ictab,{t,t0,T}]
*)
(*
The output:
{{InterpolatingFunction["<<>>"], 
   InterpolatingFunction["<<>>"]}, 
  {InterpolatingFunction["<<>>"], 
   InterpolatingFunction["<<>>"]}}
*)
